/*
2022-05-14
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#include <rp2040.h>
#include "pico_fmeter.h"

extern void F1_messung(int init);
extern void init_uart0(void);
extern uint8_t lcd_init(void);
extern void  lcd_zeile1(char *s);
extern void  lcd_zeile2(char *s);

extern int8_t   ausgabe_sperre,             // keine ser. Datenausgabe
          anzeige_funktion,                 // F+P, nur F, nur P, nur UPM
          update_LCD;                       // LCD beschriften

uint32_t SystemCoreClock  = DEF_SYSCLOCK;   // 150 MHz, leicht uebertaktet
uint32_t F_xtal = MHZ_12;                   // onboard default

double f_ref_korrektur;                     // +/- 20 ppm per R-trimm
uint32_t ms_counter;                        // fuer freie Verwedung
volatile uint8_t ms_flag, adc_flag;         // ms-flags



void kurz_warten(void)    // fuer kurze Verzoegerungen
{
  volatile int i = 10000;
  while(i--);
}  


// XOSC aktivieren, SystemCoreClock setzen und Taktsignale aktivieren
void init_clock(void)
{
volatile int i;
  VREG_AND_CHIP_RESET-> VREG = 15<<VREG_AND_CHIP_RESET_VREG_VSEL_Pos | VREG_AND_CHIP_RESET_VREG_EN_Msk;
  kurz_warten();       // Vreg abwarten

  XOSC->CTRL     = (XOSC_CTRL_FREQ_RANGE_1_15MHZ << XOSC_CTRL_FREQ_RANGE_Pos);
  XOSC->STARTUP  = 100;             // ca. 2 ms bei 12 MHz
  XOSC_SET->CTRL = (XOSC_CTRL_ENABLE_ENABLE << XOSC_CTRL_ENABLE_Pos);
  while (!(XOSC->STATUS & XOSC_STATUS_STABLE_Msk));
  
  RESETS_CLR->RESET = RESETS_RESET_pll_sys_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pll_sys_Msk));

// PLL: 12 MHz * 50 = 600 MHz
  PLL_SYS->CS = (1 << PLL_SYS_CS_REFDIV_Pos);
  PLL_SYS->FBDIV_INT = (SystemCoreClock*PLL_DIV1*PLL_DIV2)/F_xtal;  // typ. 600 MHz
// 600 MHz / 4 = 150 MHz
  PLL_SYS->PRIM = (PLL_DIV1 << PLL_SYS_PRIM_POSTDIV1_Pos) | 
                  (PLL_DIV2 << PLL_SYS_PRIM_POSTDIV2_Pos);

  PLL_SYS_CLR->PWR = PLL_SYS_PWR_VCOPD_Msk | PLL_SYS_PWR_PD_Msk;
  while (!(PLL_SYS->CS & PLL_SYS_CS_LOCK_Msk));

  PLL_SYS_CLR->PWR = PLL_SYS_PWR_POSTDIVPD_Msk;

  CLOCKS->CLK_REF_CTRL = (CLOCKS_CLK_REF_CTRL_SRC_xosc_clksrc << CLOCKS_CLK_REF_CTRL_SRC_Pos);
  CLOCKS->CLK_SYS_CTRL = (CLOCKS_CLK_SYS_CTRL_AUXSRC_clksrc_pll_sys << CLOCKS_CLK_SYS_CTRL_AUXSRC_Pos);
  CLOCKS_SET->CLK_SYS_CTRL = (CLOCKS_CLK_SYS_CTRL_SRC_clksrc_clk_sys_aux << CLOCKS_CLK_SYS_CTRL_SRC_Pos);

  CLOCKS->CLK_PERI_CTRL = CLOCKS_CLK_PERI_CTRL_ENABLE_Msk |
      (CLOCKS_CLK_PERI_CTRL_AUXSRC_clk_sys << CLOCKS_CLK_PERI_CTRL_AUXSRC_Pos);

  CLOCKS->CLK_ADC_CTRL = CLOCKS_CLK_ADC_CTRL_ENABLE_Msk |
      (CLOCKS_CLK_ADC_CTRL_AUXSRC_xosc_clksrc << CLOCKS_CLK_ADC_CTRL_AUXSRC_Pos);

// enable GPIO
  RESETS_CLR->RESET = RESETS_RESET_io_bank0_Msk | 
                      RESETS_RESET_pads_bank0_Msk;
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));
}

// GPIO14 auf '0' testen und ggf. 10 MHz Referenz verwenden
void teste_10MHz(void)
{
  IO_BANK0->GPIO14_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  PADS_BANK0_SET->GPIO14 = PADS_BANK0_GPIO0_PUE_Msk;   // pullup = 1
  PADS_BANK0_CLR->GPIO14 = PADS_BANK0_GPIO0_PDE_Msk;   // pulldon = 0
  kurz_warten();
  if(!(SIO->GPIO_IN & 1 << 14)) {
    F_xtal = MHZ_10;
    PLL_SYS->FBDIV_INT = (SystemCoreClock*PLL_DIV1*PLL_DIV2)/F_xtal;  // typ. 600 MHz
// 600 MHz / 4 = 150 MHz
    PLL_SYS_CLR->PWR = PLL_SYS_PWR_VCOPD_Msk | PLL_SYS_PWR_PD_Msk;
    while (!(PLL_SYS->CS & PLL_SYS_CS_LOCK_Msk));
  }
}

// GPIO15 auf '0' testen und ggf. SystemCoreClock verdoppeln
void teste_turbo(void)
{
  IO_BANK0->GPIO15_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  PADS_BANK0_SET->GPIO15 = PADS_BANK0_GPIO0_PUE_Msk;   // pullup = 1
  PADS_BANK0_CLR->GPIO15 = PADS_BANK0_GPIO0_PDE_Msk;   // pulldon = 0
  kurz_warten();
  if(!(SIO->GPIO_IN & 1 << 15)) {
    PLL_SYS->PRIM = (PLL_DIV1 << PLL_SYS_PRIM_POSTDIV1_Pos) | 
                  (PLL_DIV2/2 << PLL_SYS_PRIM_POSTDIV2_Pos);
    SystemCoreClock *= 2;
  }
}  

// alle Millisekunden aufrufen
void TIMER_IRQ_0_IRQHandler(void)
{
  TIMER->INTR = TIMER_INTR_ALARM_0_Msk;
  TIMER->ALARM0 = TIMER->TIMELR + MS_TEILER;
  ms_counter++;
  ms_flag = 1;
  adc_flag = 1;
}

// 64 bit Timer mit 1 us Takt und loeschen
void init_timer(void)
{
  RESETS_CLR->RESET = RESETS_RESET_timer_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_timer_Msk));

  NVIC_EnableIRQ(TIMER_IRQ_0_IRQn);
  NVIC_SetPriority(TIMER_IRQ_0_IRQn,3);
  WATCHDOG->TICK = WATCHDOG_TICK_ENABLE_Msk | (F_xtal / MHZ_FAKTOR);
  TIMER->TIMELW = 0;
  TIMER->TIMEHW = 0;
  TIMER->ALARM0 = TIMER->TIMELR + MS_TEILER;
  TIMER->INTE = TIMER_INTE_ALARM_0_Msk;
  
}

void warten_ms(uint32_t ms)
{
  do {
    while(!ms_flag);
    ms_flag = 0;
  } while(--ms);
} 


void init_adc0(void)    // ADC-Kanal0 dauernd wandeln
{
  RESETS_CLR->RESET = RESETS_RESET_adc_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_adc_Msk));
  ADC->CS = ADC_CS_START_MANY_Msk | ADC_CS_EN_Msk;
  PADS_BANK0->GPIO26 = PADS_BANK0_GPIO0_OD_Msk;   // output disable
}

uint32_t get_adc0(void) // aktuellen Wert lesen
{
  return(ADC->RESULT);
}  

// auf externen Trimmer testen und ggf. zur Frequenzkorrektur verwenden
void teste_abgleich(void)
{
static int32_t summe = ADC_MAX/2 * ADC_FILTER;  // startwert
static uint32_t init, poti_vorhanden;           // test R-trimm

  if(!init) {
    PADS_BANK0_SET->GPIO26 = PADS_BANK0_GPIO0_PUE_Msk;   // pullup = 1
    kurz_warten();
    if(get_adc0() <= 4000)poti_vorhanden = 1;
    PADS_BANK0_CLR->GPIO26 = PADS_BANK0_GPIO0_PUE_Msk;   // pullup = 0
    PADS_BANK0_SET->GPIO26 = PADS_BANK0_GPIO0_PDE_Msk;   // pulldown = 1
    kurz_warten();
    if(get_adc0() >= 100)poti_vorhanden = 1;
    PADS_BANK0_CLR->GPIO26 = PADS_BANK0_GPIO0_PDE_Msk;   // pulldown = 0
    init = 1;
  }  

  if(adc_flag && poti_vorhanden) {
    adc_flag = 0;
    summe -= summe/ADC_FILTER;
    summe += (get_adc0() - ADC_MAX/2);
// Korrekturwert von +/- 2047 auf +/- 20,47 ppm von SystemCoreClock abbilden
    f_ref_korrektur = (double)SystemCoreClock * (double)summe/ADC_FILTER*1e-8;;
  }
}  



int main()
{
  init_clock();
  teste_10MHz();            // GPIO14 == 0: F_Xtal = 10 MHz
  teste_turbo();            // GPIO15 == 0: SytemCoreClock *= 2
  init_timer();
  init_adc0();
  init_uart0();
  lcd_init();
  lcd_zeile1("mino Pico-FMeter");
  lcd_zeile2("V1.1  2023-03-04");
  ausgabe_sperre = 1;
  F1_messung(1);            // init
  while(ausgabe_sperre && TIMER->TIMELR <= 2000000) {
    F1_messung(0);          // 2 s abwarten
  }  
  ausgabe_sperre = 0;
  update_LCD = 1;           // anzeige freigeben

  while(1) {
    F1_messung(0);          // run
    teste_abgleich();
  }  
}


